#include <dirent.h>
#include <stdio.h>

#include "config.h"

#include "engine.h"
#include "gui.h"
#include "ruleset_wrapper.h"

#include "colors.h"
#include "game.h"
#include "map.h"
#include "nation.h"
#include "savegame.h"
#include "srv_main.h"
#include "terrain.h"

// We'll store the information on which savegames there are in a list
// containing the following structs:
typedef struct
{
	char*	filename;
	int		year;
} savegame_information_t;

// Create the list type and implementation
#define SPECLIST_TAG savegame_information
#define SPECLIST_TYPE savegame_information_t
#include "speclist.h"

// Variables to hold the information on the savegames we are displaying
struct savegame_information_list savegame_information;
gboolean g_game_loaded = FALSE;

engine_settings g_settings =
{
	TRUE,
	TRUE,
	TRUE,
	SHOW_ALL_PLAYERS
};

// The pixmal that the engine draws on.
// It is created in load_savegame_directory()
GdkPixmap* g_pmap_pixmap = NULL;

// Resources necessary to draw the map, these are loaded in initialize()
GdkPixbuf* cities_pixbuf = NULL;
GdkPixbuf* roads_pixbuf = NULL;
GdkPixbuf* tiles_pixbuf = NULL;
GdkPixbuf* units_pixbuf = NULL;

PangoFontDescription*	city_productions_font = NULL;
PangoFontDescription*	main_font = NULL;

int calc_terrain_sprite_xpos(struct tile* ptile);
int calc_terrain_sprite_xpos_simple(struct tile* ptile);
int compare_savegame_information_t(const void* one, const void* two);
void determine_special_sprite(int special, struct tile* ptile, int* psprite_xpos, int* psprite_ypos, GdkPixbuf** pppixbuf);
void draw_city(GdkDrawable* pdrawable, int xpos, int ypos, struct city *pcity);
void draw_city_description(GdkDrawable* pdrawable, int xpos, int ypos, struct city *pcity);
void draw_map(void);
void draw_player_flag(GdkDrawable* pdrawable, int xpos, int ypos, int player);
void draw_sprite(GdkDrawable* pdrawable, int xpos, int ypos, GdkPixbuf* ppixbuf, int sprite_xpos, int sprite_ypos);
void draw_tile(GdkDrawable* pdrawable, int xpos, int ypos, struct tile *ptile);
void draw_tile_occupants(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile);
void draw_tile_specials(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile);
void draw_tile_terrain(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile);
void draw_tile_transportation(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile, int type);
void draw_tile_unknown_edges(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile);
void draw_units(GdkDrawable* pdrawable, int xpos, int ypos, struct unit_list *punit_list);

void free_savegame_information(void);

void get_city_mapview_name_and_growth(struct city *pcity,
									  char *name_buffer,
									  size_t name_buffer_len,
									  char *growth_buffer,
									  size_t growth_buffer_len,
									  enum color_std *growth_color);
void get_city_mapview_production(struct city* pcity,
								 char* buffer,
								 size_t buffer_len);
void gtk_draw_shadowed_string(GdkDrawable* drawable,
						      GdkGC* black_gc,
						      GdkGC* white_gc,
						      gint x,
						      gint y,
						      PangoLayout* layout);

bool load_savegame(struct section_file* savegame, GError** pperror);
bool load_savegame_from_filename(char* filename, GError** pperror);

int calc_terrain_sprite_xpos(struct tile* ptile)
{
	int deduct, xpos;
	
	// If this is the only tile with this terrain around here, use the sprite
	// at position 15. Depending on where similar terrain is, deduct from
	// this 15 to get the accurate sprite.

	xpos = 15;
	
	cardinal_adjc_dir_iterate(ptile, ptile_iter, dir_iter)
	{
		deduct = 0;
		
		switch (dir_iter)
		{
		case DIR8_NORTH:
			deduct = 1;
			break;
		case DIR8_EAST:
			deduct = 2;
			break;
		case DIR8_SOUTH:
			deduct = 4;
			break;
		case DIR8_WEST:
			deduct = 8;
			break;
		default:
			printf("ERROR: unknown direction: %d\n", dir_iter);
			break;
		}

		if (ptile_iter->terrain == ptile->terrain)
		{
			xpos -= deduct;
		}
	} cardinal_adjc_dir_iterate_end;

	return xpos;
}

int calc_terrain_sprite_xpos_simple(struct tile* ptile)
{
	gboolean left = FALSE;
	gboolean right = FALSE;
	
	cardinal_adjc_dir_iterate(ptile, ptile_iter, dir_iter)
	{
		switch (dir_iter)
		{
		case DIR8_NORTH:
			break;
		case DIR8_EAST:
			if (ptile_iter->terrain == ptile->terrain)
			{
				right = TRUE;
			}
			break;
		case DIR8_SOUTH:
			break;
		case DIR8_WEST:
			if (ptile_iter->terrain == ptile->terrain)
			{
				left = TRUE;
			}
			break;
		default:
			printf("ERROR: unknown direction: %d\n", dir_iter);
			break;
		}
	} cardinal_adjc_dir_iterate_end;

	if (!left && !right)
	{
		return 0;
	}
	else if (left && right)
	{
		return 2;
	}
	else if (left)
	{
		return 3;
	}
	else
	{
		return 1;
	}
}

int compare_savegame_information_t(const void* one, const void* two)
{
	savegame_information_t* pone = *((savegame_information_t**) one);
	savegame_information_t* ptwo = *((savegame_information_t**) two);

	if (pone->year < ptwo->year)
	{
		return -1;
	}
	else if (pone->year > ptwo->year)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

void determine_special_sprite(int special, struct tile* ptile, int* psprite_xpos, int* psprite_ypos, GdkPixbuf** pppixbuf)
{
	int deduct, xpos;
		
	switch (special)
	{
	case S_NO_SPECIAL:
		*pppixbuf = NULL;
		break;
	case S_SPECIAL_1:
		*pppixbuf = tiles_pixbuf;
		switch (ptile->terrain)
		{
		case 0: // glacier
			*psprite_xpos = 19;
			*psprite_ypos = 5;
			break;
		case 1: // desert
			*psprite_xpos = 1;
			*psprite_ypos = 11;
			break;
		case 2: // forest
			*psprite_xpos = 16;
			*psprite_ypos = 7;
			break;
		case 3: // grassland
			*psprite_xpos = 3;
			*psprite_ypos = 11;
			break;
		case 4: // hills
			*psprite_xpos = 4;
			*psprite_ypos = 11;
			break;
		case 5: // jungle
			*psprite_xpos = 5;
			*psprite_ypos = 11;
			break;
		case 6: // mountains
			*psprite_xpos = 6;
			*psprite_ypos = 11;
			break;
		case 7: // ocean
			*psprite_xpos = 7;
			*psprite_ypos = 11;
			break;
		case 8: // plains
			*psprite_xpos = 17;
			*psprite_ypos = 7;
			break;
		case 9: // swamp
			*psprite_xpos = 18;
			*psprite_ypos = 5;
			break;
		case 10: // tundra
			*psprite_xpos = 11;
			*psprite_ypos = 11;
			break;
		default:
			printf("ERROR: unknown terrain: %d\n", ptile->terrain);
			*pppixbuf = NULL;
			break;
		}
		break;
	case S_ROAD:
		*pppixbuf = NULL;
		break;
	case S_IRRIGATION:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 8;
		*psprite_ypos = 12;
		break;
	case S_RAILROAD:
		*pppixbuf = NULL;
		break;
	case S_MINE:
		*pppixbuf = tiles_pixbuf;
		if (ptile->terrain == 0 || ptile->terrain == 1)
		{
			// desert or arctic, use oil mine
			*psprite_xpos = 10;
			*psprite_ypos = 12;
		}
		else
		{
			// regular mine
			*psprite_xpos = 9;
			*psprite_ypos = 12;
		}
		break;
	case S_POLLUTION:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 11;
		*psprite_ypos = 12;
		break;
	case S_HUT:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 14;
		*psprite_ypos = 12;
		break;
	case S_FORTRESS:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 15;
		*psprite_ypos = 12;
		break;
	case S_SPECIAL_2:
		*pppixbuf = tiles_pixbuf;
		switch (ptile->terrain)
		{
		case 0: // glacier
			*psprite_xpos = 10;
			*psprite_ypos = 11;
			break;
		case 1: // desert
			*psprite_xpos = 18;
			*psprite_ypos = 7;
			break;
		case 2: // forest
			*psprite_xpos = 16;
			*psprite_ypos = 7;
			break;
		case 3: // grassland
			*psprite_xpos = 3;
			*psprite_ypos = 11;
			break;
		case 4: // hills
			*psprite_xpos = 19;
			*psprite_ypos = 7;
			break;
		case 5: // jungle
			*psprite_xpos = 16;
			*psprite_ypos = 6;
			break;
		case 6: // mountains
			*psprite_xpos = 17;
			*psprite_ypos = 6;
			break;
		case 7: // ocean
			*psprite_xpos = 18;
			*psprite_ypos = 6;
			break;
		case 8: // plains
			*psprite_xpos = 19;
			*psprite_ypos = 6;
			break;
		case 9: // swamp
			*psprite_xpos = 16;
			*psprite_ypos = 5;
			break;
		case 10: // tundra
			*psprite_xpos = 17;
			*psprite_ypos = 5;
			break;
		default:
			printf("ERROR: unknown terrain: %d\n", ptile->terrain);
			*pppixbuf = NULL;
			break;
		}
		break;
	case S_RIVER:
		// If this is the only tile with this terrain around here, use the sprite
		// at position 15. Depending on where similar terrain is, deduct from
		// this 15 to get the accurate sprite.
	
		xpos = 15;
		
		cardinal_adjc_dir_iterate(ptile, ptile_iter, dir_iter)
		{
			deduct = 0;
			
			switch (dir_iter)
			{
			case DIR8_NORTH:
				deduct = 1;
				break;
			case DIR8_EAST:
				deduct = 2;
				break;
			case DIR8_SOUTH:
				deduct = 4;
				break;
			case DIR8_WEST:
				deduct = 8;
				break;
			default:
				printf("ERROR: unknown direction: %d\n", dir_iter);
				break;
			}

			// If ptile_iter does not have a river and is not an ocean
			// don't draw a river toward there.
			if ((ptile_iter->special & S_RIVER) == 0 &&
				ptile_iter->terrain != 7)
			{
				xpos -= deduct;
			}
		} cardinal_adjc_dir_iterate_end;

		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = xpos;
		*psprite_ypos = 18;

		break;
	case S_FARMLAND:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 7;
		*psprite_ypos = 12;
		break;
	case S_AIRBASE:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 16;
		*psprite_ypos = 13;
		break;
	case S_FALLOUT:
		*pppixbuf = tiles_pixbuf;
		*psprite_xpos = 18;
		*psprite_ypos = 13;
		break;
	default:
		printf("ERROR: unknown special: %d\n", special);
		*pppixbuf = NULL;
		break;
	}
}

void draw_city(GdkDrawable* pdrawable, int xpos, int ypos, struct city *pcity)
{
	// dependent on the nation, choose the appropriate city style
	int city_pixbuf_sprite_ypos;

	struct nation_type* pnation = get_nation_by_plrid(pcity->owner);
	
	city_pixbuf_sprite_ypos = pnation->city_style + 1;

	// draw the flag
	draw_player_flag(pdrawable, xpos, ypos, pcity->owner);
	
	// draw the actual city sprite
	int sprite_xpos;
	
	if (pcity->size < 5)
	{
		sprite_xpos = 0;
	}
	else if (pcity->size < 10)
	{
		sprite_xpos = 1;
	}
	else
	{
		sprite_xpos = 2;
	}

	draw_sprite(pdrawable, xpos, ypos, cities_pixbuf, sprite_xpos, city_pixbuf_sprite_ypos);
	
	// draw the city size
	draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, pcity->size % 10, 14);
	
	if (pcity->size > 9)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, 9 + pcity->size / 10, 14);
	}
	
	// draw the flag in city if unit is present
	if (unit_list_size(&pcity->tile->units) > 0)
	{
		draw_sprite(pdrawable, xpos, ypos, cities_pixbuf, 4, city_pixbuf_sprite_ypos);
	}
	
	if (city_got_citywalls(pcity))
	{
		draw_sprite(pdrawable, xpos, ypos, cities_pixbuf, 3, city_pixbuf_sprite_ypos);
	}

	draw_city_description(pdrawable, xpos, ypos + 1, pcity);
}

/* show_city_desc from mapview.c */
void draw_city_description(GdkDrawable* pdrawable, int xpos, int ypos, struct city *pcity)
{
	static char buffer[512], buffer2[32];
	PangoRectangle rect, rect2;
	enum color_std color;
	int extra_width = 0;
	static PangoLayout *layout;
	
	if (!layout) {
		layout = pango_layout_new(gdk_pango_context_get());
	}

	int pixel_xpos = xpos * 30 + 30 / 2;
	int pixel_ypos = ypos * 30;
	
	if (g_settings.draw_city_names)
	{
		get_city_mapview_name_and_growth(pcity, buffer, sizeof(buffer), buffer2, sizeof(buffer2), &color);
		
		pango_layout_set_font_description(layout, main_font);

		if (buffer2[0] != '\0') {
			/* HACK: put a character's worth of space between the two strings. */
			pango_layout_set_text(layout, "M", -1);
			pango_layout_get_pixel_extents(layout, &rect, NULL);
			extra_width = rect.width;
		}

		pango_layout_set_text(layout, buffer, -1);
		pango_layout_get_pixel_extents(layout, &rect, NULL);
		rect.width += extra_width;
		
		if (g_settings.draw_city_growth && (
				g_settings.current_player == SHOW_ALL_PLAYERS ||
				pcity->owner == g_settings.current_player))
		{
			/* We need to know the size of the growth text before
			drawing anything. */
			pango_layout_set_font_description(layout, city_productions_font);
			pango_layout_set_text(layout, buffer2, -1);
			pango_layout_get_pixel_extents(layout, &rect2, NULL);
			
			/* Now return the layout to its previous state. */
			pango_layout_set_font_description(layout, main_font);
			pango_layout_set_text(layout, buffer, -1);
		}
		else
		{
			rect2.width = 0;
		}
	
		gtk_draw_shadowed_string(
			pdrawable,
			get_mapcanvas_widget()->style->black_gc,
			get_mapcanvas_widget()->style->white_gc,
			pixel_xpos - (rect.width + rect2.width) / 2,
			pixel_ypos + PANGO_ASCENT(rect), layout);
		
		if (g_settings.draw_city_growth && (
				g_settings.current_player == SHOW_ALL_PLAYERS ||
				pcity->owner == g_settings.current_player))
		{
			pango_layout_set_font_description(layout, city_productions_font);
			pango_layout_set_text(layout, buffer2, -1);

			// gdk_gc_set_foreground(civ_gc, colors_standard[color]);

			gtk_draw_shadowed_string(
				pdrawable,
				get_mapcanvas_widget()->style->black_gc,
				get_mapcanvas_widget()->style->white_gc,
				pixel_xpos - (rect.width + rect2.width) / 2 + rect.width,
				pixel_ypos + PANGO_ASCENT(rect) + rect.height / 2 - rect2.height / 2,
				layout);
		}

		pixel_ypos += rect.height + 3;
	}
	
	if (g_settings.draw_city_production && (
			g_settings.current_player == SHOW_ALL_PLAYERS ||
			pcity->owner == g_settings.current_player))
	{
		get_city_mapview_production(pcity, buffer, sizeof(buffer));
		
		pango_layout_set_font_description(layout, city_productions_font);
		pango_layout_set_text(layout, buffer, -1);
		pango_layout_get_pixel_extents(layout, &rect, NULL);

		gtk_draw_shadowed_string(
			pdrawable,
			get_mapcanvas_widget()->style->black_gc,
			get_mapcanvas_widget()->style->white_gc,
			pixel_xpos - rect.width / 2,
			pixel_ypos + PANGO_ASCENT(rect),
			layout);
	}
}

void draw_map(void)
{
	gdk_draw_rectangle(
		g_pmap_pixmap,
		get_mapcanvas_widget()->style->black_gc,
		TRUE,
		0,
		0,
		map.xsize * 30,
		map.ysize * 30);

	int x, y;

	for (y = 0; y < map.ysize; y++)
	{
		for (x = 0; x < map.xsize; x++)
		{
			draw_tile(g_pmap_pixmap, x, y, native_pos_to_tile(x, y));
		}
	}

	for (y = 0; y < map.ysize; y++)
	{
		for (x = 0; x < map.xsize; x++)
		{
			draw_tile_occupants(g_pmap_pixmap, x, y, native_pos_to_tile(x, y));
		}
	}
}

void draw_player_flag(GdkDrawable* pdrawable, int xpos, int ypos, int player)
{
	struct player* pplayer = &game.players[player];
	
	GdkPixbuf* ppixbuf = get_player_flag_pixbuf(pplayer);
	
	draw_sprite(pdrawable, xpos, ypos, ppixbuf, 0, 0);
}

void draw_sprite(GdkDrawable* pdrawable, int xpos, int ypos, GdkPixbuf* ppixbuf, int sprite_xpos, int sprite_ypos)
{
	gdk_draw_pixbuf(
		pdrawable,
		NULL,
		ppixbuf,
		sprite_xpos * 30,
		sprite_ypos * 30,
		xpos * 30,
		ypos * 30,
		30,
		30,
		GDK_RGB_DITHER_NORMAL,
		0,
		0);
}

void draw_tile(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile)
{
	enum known_type known;
	
	if (g_settings.current_player == SHOW_ALL_PLAYERS)
	{
		known = TILE_KNOWN;
	}
	else
	{
		known = map_get_known(ptile, &game.players[g_settings.current_player]);
	}
	
	if (known == TILE_UNKNOWN)
	{
		return;
	}

	draw_tile_terrain(pdrawable, xpos, ypos, ptile);
	draw_tile_specials(pdrawable, xpos, ypos, ptile);
	draw_tile_unknown_edges(pdrawable, xpos, ypos, ptile);

	if (known == TILE_KNOWN_FOGGED)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, 17, 13);
	}
}

void draw_tile_occupants(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile)
{
	enum known_type known;
	
	if (g_settings.current_player == SHOW_ALL_PLAYERS)
	{
		known = TILE_KNOWN;
	}
	else
	{
		known = map_get_known(ptile, &game.players[g_settings.current_player]);
	}
	
	if (known == TILE_UNKNOWN)
	{
		return;
	}

	// Draw units and city. If the tile contains both, only draw the city.
	if (ptile->city != NULL)
	{
		draw_city(pdrawable, xpos, ypos, ptile->city);
	}
	else if (unit_list_size(&ptile->units) > 0 && known == TILE_KNOWN)
	{
		draw_units(pdrawable, xpos, ypos, &ptile->units);
	}

	if (known == TILE_KNOWN_FOGGED)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, 17, 13);
	}
}

void draw_tile_specials(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile)
{
	if (ptile->special == S_NO_SPECIAL)
	{
		return;
	}
	
	int order_count = 13;
	int special_order[] = {
		S_RIVER,
		S_IRRIGATION,
		S_FARMLAND,
		S_ROAD,
		S_RAILROAD,
		S_SPECIAL_1,
		S_SPECIAL_2,
		S_HUT,
		S_MINE,
		S_FORTRESS,
		S_AIRBASE,
		S_POLLUTION,
		S_FALLOUT
	};

	int i;	
	for (i = 0; i < order_count; i++)
	{
		int special = special_order[i];
		
		if ((ptile->special & special) != 0)
		{
			if (special == S_ROAD || special == S_RAILROAD)
			{
				draw_tile_transportation(pdrawable, xpos, ypos, ptile, special);
			}
			else
			{
				int sprite_xpos;
				int sprite_ypos;
				GdkPixbuf* ppixbuf;
				
				determine_special_sprite(
					special,
					ptile,
					&sprite_xpos,
					&sprite_ypos,
					&ppixbuf);
				
				if (ppixbuf != NULL)
				{
					draw_sprite(pdrawable, xpos, ypos, ppixbuf, sprite_xpos, sprite_ypos);
				}
			}
		}
	}
}

void draw_tile_terrain(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile)
{
	int sprite_xpos, sprite_ypos;
	
	switch (ptile->terrain)
	{
	case 0: // glacier
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 3;
		break;
	case 1: // desert
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 2;
		break;
	case 2: // forest
		sprite_xpos = 4 + calc_terrain_sprite_xpos_simple(ptile);
		sprite_ypos = 10;
		break;
	case 3: // grassland
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 1;
		break;
	case 4: // hills
		sprite_xpos = 0 + calc_terrain_sprite_xpos_simple(ptile);
		sprite_ypos = 10;
		break;
	case 5: // jungle
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 4;
		break;
	case 6: // mountains
		sprite_xpos = 8 + calc_terrain_sprite_xpos_simple(ptile);
		sprite_ypos = 10;
		break;
	case 7: // ocean
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 9;
		break;
	case 8: // plains
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 5;
		break;
	case 9: // swamp
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 6;
		break;
	case 10: // tundra
		sprite_xpos = calc_terrain_sprite_xpos(ptile);
		sprite_ypos = 7;
		break;
	default:
		printf("ERROR: unknown terrain: %d\n", ptile->terrain);
		return;
	}

	draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, sprite_xpos, sprite_ypos);
}

void draw_tile_transportation(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile, int type)
{
	int card_xpos, cross_xpos;
	
	card_xpos = 15;
	cross_xpos = 15;

	adjc_dir_iterate(ptile, ptile_iter, dir_iter)
	{
		if ((ptile_iter->special & type) != type)
		{
			switch (dir_iter)
			{
			case DIR8_NORTHWEST:
				cross_xpos -= 8;
				break;
			case DIR8_NORTH:
				card_xpos -= 1;
				break;
			case DIR8_NORTHEAST:
				cross_xpos -= 1;
				break;
			case DIR8_EAST:
				card_xpos -= 2;
				break;
			case DIR8_SOUTHEAST:
				cross_xpos -= 2;
				break;
			case DIR8_SOUTH:
				card_xpos -= 4;
				break;
			case DIR8_SOUTHWEST:
				cross_xpos -= 4;
				break;
			case DIR8_WEST:
				card_xpos -= 8;
				break;
			default:
				printf("ERROR: unknown direction: %d\n", dir_iter);
				return;
			}
		}
	} adjc_dir_iterate_end;

	int sprite_ypos;
	
	switch (type)
	{
	case S_ROAD:
		sprite_ypos = 0;
		break;
	case S_RAILROAD:
		sprite_ypos = 4;
		break;
	default:
		printf("ERROR: transportation type: %d\n", type);
		return;
	}

	// Only draw card sprite 0 if there is no cross sprite
	if (card_xpos != 0 || (card_xpos == 0 && cross_xpos == 0))
	{
		draw_sprite(pdrawable, xpos, ypos, roads_pixbuf, card_xpos, sprite_ypos);
	}
	
	draw_sprite(pdrawable, xpos, ypos, roads_pixbuf, cross_xpos, sprite_ypos + 1);
}

void draw_tile_unknown_edges(GdkDrawable* pdrawable, int xpos, int ypos, struct tile* ptile)
{
	if (g_settings.current_player == SHOW_ALL_PLAYERS)
	{
		return;
	}
	
	int deduct, sprite_xpos;
	
	// If this is the only tile that is known to the current player, use the
	// sprite at position 15. Depending on where similar terrain is, deduct
	// from this 15 to get the accurate sprite.
	
	sprite_xpos = 15;
	
	cardinal_adjc_dir_iterate(ptile, ptile_iter, dir_iter)
	{
		deduct = 0;
		
		switch (dir_iter)
		{
		case DIR8_NORTH:
			deduct = 1;
			break;
		case DIR8_EAST:
			deduct = 2;
			break;
		case DIR8_SOUTH:
			deduct = 4;
			break;
		case DIR8_WEST:
			deduct = 8;
			break;
		default:
			printf("ERROR: unknown direction: %d\n", dir_iter);
			break;
		}

		if (map_get_known(
				ptile_iter,
				&game.players[g_settings.current_player]) != TILE_UNKNOWN)
		{
			sprite_xpos -= deduct;
		}
	} cardinal_adjc_dir_iterate_end;
	
	draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, sprite_xpos, 13);
}

void draw_units(GdkDrawable* pdrawable, int xpos, int ypos, struct unit_list *punit_list)
{
	int unit_count = unit_list_size(punit_list);
	
	if (unit_count == 0)
	{
		return;
	}

	struct unit* pdraw_unit = NULL;
	int level_found = -1;
	
	unit_list_iterate_safe(*punit_list, punit)
	{
		/* Draw levels:
		 *  - 6 ground unit carriers on top
		 *  - 5 air unit carriers
		 *  - 4 air/heli units
		 *  - 3 water units
		 *  - 2 ground units
		 *  - 1 the rest
		 */
		 
		if (is_ground_units_transport(punit))
		{
			pdraw_unit = punit;
			level_found = 6;
		}
		else if (is_air_units_transport(punit) && level_found < 5)
		{
			pdraw_unit = punit;
			level_found = 5;
		}
		else if ((is_air_unit(punit) || is_heli_unit(punit)) && level_found < 4)
		{
			pdraw_unit = punit;
			level_found = 4;
		}
		else if (is_sailing_unit(punit) && level_found < 3)
		{
			pdraw_unit = punit;
			level_found = 3;
		}
		else if (is_ground_unit(punit) && level_found < 2)
		{
			pdraw_unit = punit;
			level_found = 2;
		}
		else if (level_found < 1)
		{
			pdraw_unit = punit;
			level_found = 1;
		}
	} unit_list_iterate_safe_end;

	int sprite_xpos, sprite_ypos;
	
	// Get this from the ruleset ...
	switch (pdraw_unit->type)
	{
		case 0: // unit_settlers
			sprite_xpos = 4;
			sprite_ypos = 1;
			break;
		case 1: // unit_engineers
			sprite_xpos = 2;
			sprite_ypos = 2;
			break;
		case 2: // unit_warriors
			sprite_xpos = 18;
			sprite_ypos = 0;
			break;
		case 3: // unit_phalanx
			sprite_xpos = 1;
			sprite_ypos = 1;
			break;
		case 4: // unit_archers
			sprite_xpos = 8;
			sprite_ypos = 1;
			break;
		case 5: // unit_legion
			sprite_xpos = 16;
			sprite_ypos = 0;
			break;
		case 6: // unit_pikemen
			sprite_xpos = 17;
			sprite_ypos = 1;
			break;
		case 7: // unit_musketeers
			sprite_xpos = 19;
			sprite_ypos = 0;
			break;
		case 8: // unit_fanatics
			sprite_xpos = 12;
			sprite_ypos = 2;
			break;
		case 9: // unit_partisan
			sprite_xpos = 16;
			sprite_ypos = 1;
			break;
		case 10: // unit_alpine_troops
			sprite_xpos = 5;
			sprite_ypos = 2;
			break;
		case 11: // unit_riflemen
			sprite_xpos = 2;
			sprite_ypos = 1;
			break;
		case 12: // unit_marines
			sprite_xpos = 0;
			sprite_ypos = 2;
			break;
		case 13: // unit_paratroopers
			sprite_xpos = 9;
			sprite_ypos = 2;
			break;
		case 14: // unit_mech_inf
			sprite_xpos = 17;
			sprite_ypos = 0;
			break;
		case 15: // unit_horsemen
			sprite_xpos = 8;
			sprite_ypos = 0;
			break;
		case 16: // unit_chariot
			sprite_xpos = 9;
			sprite_ypos = 0;
			break;
		case 17: // unit_elephants
			sprite_xpos = 10;
			sprite_ypos = 2;
			break;
		case 18: // unit_crusaders
			sprite_xpos = 11;
			sprite_ypos = 2;
			break;
		case 19: // unit_knights
			sprite_xpos = 15;
			sprite_ypos = 0;
			break;
		case 20: // unit_dragoons
			sprite_xpos = 12;
			sprite_ypos = 1;
			break;
		case 21: // unit_cavalry
			sprite_xpos = 9;
			sprite_ypos = 1;
			break;
		case 22: // unit_armor
			sprite_xpos = 0;
			sprite_ypos = 0;
			break;
		case 23: // unit_catapult
			sprite_xpos = 7;
			sprite_ypos = 0;
			break;
		case 24: // unit_cannon
			sprite_xpos = 4;
			sprite_ypos = 0;
			break;
		case 25: // unit_artillery
			sprite_xpos = 3;
			sprite_ypos = 2;
			break;
		case 26: // unit_howitzer
			sprite_xpos = 1;
			sprite_ypos = 0;
			break;
		case 27: // unit_fighter
			sprite_xpos = 12;
			sprite_ypos = 0;
			break;
		case 28: // unit_bomber
			sprite_xpos = 3;
			sprite_ypos = 0;
			break;
		case 29: // unit_helicopter
			sprite_xpos = 4;
			sprite_ypos = 2;
			break;
		case 30: // unit_stealth_fighter
			sprite_xpos = 7;
			sprite_ypos = 2;
			break;
		case 31: // unit_stealth_bomber
			sprite_xpos = 6;
			sprite_ypos = 2;
			break;
		case 32: // unit_trireme
			sprite_xpos = 7;
			sprite_ypos = 1;
			break;
		case 33: // unit_caravel
			sprite_xpos = 3;
			sprite_ypos = 1;
			break;
		case 34: // unit_galleon
			sprite_xpos = 15;
			sprite_ypos = 1;
			break;
		case 35: // unit_frigate
			sprite_xpos = 13;
			sprite_ypos = 0;
			break;
		case 36: // unit_ironclad
			sprite_xpos = 14;
			sprite_ypos = 0;
			break;
		case 37: // unit_destroyer
			sprite_xpos = 11;
			sprite_ypos = 1;
			break;
		case 38: // unit_cruiser
			sprite_xpos = 10;
			sprite_ypos = 0;
			break;
		case 39: // unit_aegis_cruiser
			sprite_xpos = 8;
			sprite_ypos = 2;
			break;
		case 40: // unit_battleship
			sprite_xpos = 2;
			sprite_ypos = 0;
			break;
		case 41: // unit_submarine
			sprite_xpos = 5;
			sprite_ypos = 1;
			break;
		case 42: // unit_carrier
			sprite_xpos = 6;
			sprite_ypos = 0;
			break;
		case 43: // unit_transport
			sprite_xpos = 6;
			sprite_ypos = 1;
			break;
		case 44: // unit_cruise_missile
			sprite_xpos = 10;
			sprite_ypos = 1;
			break;
		case 45: // unit_nuclear
			sprite_xpos = 0;
			sprite_ypos = 1;
			break;
		case 46: // unit_diplomat
			sprite_xpos = 11;
			sprite_ypos = 0;
			break;
		case 47: // unit_spy
			sprite_xpos = 2;
			sprite_ypos = 1;
			break;
		case 48: // unit_caravan
			sprite_xpos = 5;
			sprite_ypos = 0;
			break;
		case 49: // unit_freight
			sprite_xpos = 14;
			sprite_ypos = 1;
			break;
		case 50: // unit_explorer
			sprite_xpos = 13;
			sprite_ypos = 1;
			break;
		case 51: // unit_barbarian_leader
			sprite_xpos = 11;
			sprite_ypos = 0;
			break;
		case 52: // unit_leader
			sprite_xpos = 13;
			sprite_ypos = 1;
			break;
		default:
			printf("ERROR: unknown unit type: %d\n", pdraw_unit->type);

			// lets just draw a warrior here :-)
			sprite_xpos = 18;
			sprite_ypos = 0;
	}
	
	draw_player_flag(pdrawable, xpos, ypos, pdraw_unit->owner);

	draw_sprite(pdrawable, xpos, ypos, units_pixbuf, sprite_xpos, sprite_ypos);
	
	if (unit_count > 1)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, 19, 9);
	}
	
	if (pdraw_unit->veteran)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, 11, 17);
	}

	/* There are 11 hitpoint sprites. If the unit has full hitpoints, the
	 * 0th one should be drawn; if it is nearly out of hitpoits, the 10th.
	 * A unit cannot be out of hitpoints, because then it is dead and will
	 * not be drawn here ;-)
	 */
	draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf,
		10 - (10 * pdraw_unit->hp) / unit_type(pdraw_unit)->hp, 16);

	switch (pdraw_unit->activity)
	{
	case ACTIVITY_MINE:
		sprite_xpos = 15;
		sprite_ypos = 11;
		break;
	case ACTIVITY_POLLUTION:
		sprite_xpos = 16;
		sprite_ypos = 11;
		break;
	case ACTIVITY_FALLOUT:
		sprite_xpos = 19;
		sprite_ypos = 12;
		break;
	case ACTIVITY_PILLAGE:
		sprite_xpos = 17;
		sprite_ypos = 12;
		break;
	case ACTIVITY_ROAD:
	case ACTIVITY_RAILROAD:
		sprite_xpos = 17;
		sprite_ypos = 11;
		break;
	case ACTIVITY_IRRIGATE:
		sprite_xpos = 18;
		sprite_ypos = 11;
		break;
	case ACTIVITY_EXPLORE:
		sprite_xpos = 19;
		sprite_ypos = 10;
		break;
	case ACTIVITY_FORTIFIED:
		sprite_xpos = 18;
		sprite_ypos = 12;
		break;
	case ACTIVITY_FORTIFYING:
		sprite_xpos = 19;
		sprite_ypos = 11;
		break;
	case ACTIVITY_AIRBASE:
		sprite_xpos = 16;
		sprite_ypos = 12;
		break;
	case ACTIVITY_SENTRY:
		sprite_xpos = 13;
		sprite_ypos = 11;
		break;
	case ACTIVITY_TRANSFORM:
		sprite_xpos = 12;
		sprite_ypos = 12;
		break;
	case ACTIVITY_FORTRESS:
	case ACTIVITY_GOTO:
	default:
		sprite_xpos = -1;
		sprite_ypos = -1;
		break;
	}

	if (sprite_xpos != -1)
	{
		draw_sprite(pdrawable, xpos, ypos, tiles_pixbuf, sprite_xpos, sprite_ypos);
	}
}

void free_savegame_information(void)
{
	int i, size;

	size = savegame_information_list_size(&savegame_information);

	for (i = 0; i < size; i++)
	{
		savegame_information_t* psavegame_information =
			savegame_information_list_get(&savegame_information, i);
		free(psavegame_information->filename);
		free(psavegame_information);
	}
	
	savegame_information_list_init(&savegame_information);

	if (g_pmap_pixmap != NULL)
	{
		g_object_unref(g_pmap_pixmap);
		g_pmap_pixmap = NULL;
	}
}

GdkPixmap* get_map(void)
{
	return g_pmap_pixmap;
}

int get_turn_count(void)
{
	return savegame_information_list_size(&savegame_information);
}

void get_settings(engine_settings* psettings)
{
	*psettings = g_settings;
}

gboolean goto_turn(int turn, GError** pperror)
{
	if (turn >= get_turn_count())
	{
		g_error("Using turn greater than g_savegame_count");
		return FALSE;
	}

	savegame_information_t* psavegame_information =
		savegame_information_list_get(&savegame_information, turn);

	if (!load_savegame_from_filename(psavegame_information->filename, pperror))
	{
		return FALSE;
	}

	if (g_pmap_pixmap == NULL)
	{
		g_pmap_pixmap = gdk_pixmap_new(
			get_mapcanvas_widget()->window,
			map.xsize * 30,
			map.ysize * 30,
			-1);
	}

	draw_map();

	return TRUE;
}

// get_city_mapview_name_and_growth from mapview_common.c
void get_city_mapview_name_and_growth(struct city *pcity,
									  char *name_buffer,
									  size_t name_buffer_len,
									  char *growth_buffer,
									  size_t growth_buffer_len,
									  enum color_std *growth_color)
{
	if (!g_settings.draw_city_names)
	{
		name_buffer[0] = '\0';
		growth_buffer[0] = '\0';
		*growth_color = COLOR_STD_WHITE;
		return;
	}
	
	snprintf(name_buffer, name_buffer_len, pcity->name);
	
	if (g_settings.draw_city_growth) // && pcity->owner == game.player_idx)
	{
		int turns = city_turns_to_grow(pcity);
		
		if (turns == 0)
		{
			snprintf(growth_buffer, growth_buffer_len, "X");
		}
		else if (turns == FC_INFINITY)
		{
			snprintf(growth_buffer, growth_buffer_len, "-");
		}
		else
		{
			/* Negative turns means we're shrinking, but that's handled
			down below. */
			snprintf(growth_buffer, growth_buffer_len, "%d", abs(turns));
		}
		
		if (turns <= 0)
		{
			/* A blocked or shrinking city has its growth status shown in red. */
			*growth_color = COLOR_STD_RED;
		} else {
			*growth_color = COLOR_STD_WHITE;
		}
	}
	else
	{
		growth_buffer[0] = '\0';
		*growth_color = COLOR_STD_WHITE;
	}
}

// From mapview_common.c
void get_city_mapview_production(struct city* pcity,
								 char* buffer,
								 size_t buffer_len)
{
	int turns = city_turns_to_build(
		pcity,
		pcity->currently_building,
		pcity->is_building_unit,
		TRUE);
	
	if (pcity->is_building_unit)
	{
		struct unit_type *punit_type = get_unit_type(pcity->currently_building);
		
		if (turns < 999)
		{
			snprintf(buffer, buffer_len, "%s %d",
			punit_type->name, turns);
		}
		else
		{
			snprintf(buffer, buffer_len, "%s -", punit_type->name);
		}
	}
	else
	{
		struct impr_type *pimprovement_type =
			get_improvement_type(pcity->currently_building);

		if (get_current_construction_bonus(pcity, EFT_PROD_TO_GOLD) > 0)
		{
			snprintf(buffer, buffer_len, "%s", pimprovement_type->name);
		}
		else if (turns < 999)
		{
			snprintf(buffer, buffer_len, "%s %d", pimprovement_type->name, turns);
		} else {
			snprintf(buffer, buffer_len, "%s -", pimprovement_type->name);
		}
	}
}

// From graphics.c
void gtk_draw_shadowed_string(GdkDrawable* drawable,
						      GdkGC* black_gc,
						      GdkGC* white_gc,
						      gint x,
						      gint y,
						      PangoLayout* layout)
{
	gdk_draw_layout(drawable, black_gc, x + 1, y + 1, layout);
	gdk_draw_layout(drawable, white_gc, x, y, layout);
}

gboolean initialize_engine(GError** pperror)
{
	savegame_information_list_init(&savegame_information);

	/* Load fonts ... from gui_main.c:
	 * font names shouldn't be in spec files! */
	GtkStyle* style = gtk_rc_get_style_by_paths(gtk_settings_get_default(),
												"Freeciv*.city names",
												NULL,
												G_TYPE_NONE);

	if (!style)
	{
		style = gtk_style_new();
	}
	
	g_object_ref(style);
	
	main_font = style->font_desc;

	style = gtk_rc_get_style_by_paths(gtk_settings_get_default(),
									  "Freeciv*.city productions",
									  NULL,
									  G_TYPE_NONE);

	if (!style)
	{
		style = gtk_style_new();
	}
	
	g_object_ref(style);
	
	city_productions_font = style->font_desc;

	// Load tile pixbufs
	cities_pixbuf = gdk_pixbuf_new_from_file(
		"./data/trident/cities.png", pperror);
	if (cities_pixbuf == NULL)
	{
		return FALSE;
	}

	roads_pixbuf = gdk_pixbuf_new_from_file(
		"./data/trident/roads.png", pperror);
	if (roads_pixbuf == NULL)
	{
		return FALSE;
	}

	tiles_pixbuf = gdk_pixbuf_new_from_file(
		"./data/trident/tiles.png", pperror);
	if (tiles_pixbuf == NULL)
	{
		return FALSE;
	}

	units_pixbuf = gdk_pixbuf_new_from_file(
		"./data/trident/units.png", pperror);
	if (units_pixbuf == NULL)
	{
		return FALSE;
	}
	
	return TRUE;
}

gboolean is_game_loaded(void)
{
	return (get_turn_count() > 0);
}

bool load_savegame(struct section_file* psavegame, GError** pperror)
{
	if (g_game_loaded)
	{
		server_game_free();
	}

	game_init();

	game_load(psavegame);
	
	g_game_loaded = TRUE;

	return TRUE;
}

gboolean load_savegame_directory(char* directoryname, GError** pperror)
{
	char fullname[1024];

	char*	savegame_extention = ".sav.gz";
	int		savegame_extention_length = strlen(savegame_extention);

	free_savegame_information();

	DIR* dir = opendir(directoryname);
	struct dirent* entry;

	char* gameid = NULL;

	while (dir)
	{
		entry = readdir(dir);

		if (entry != NULL)
		{
			int length = strlen(entry->d_name);

			if (length > savegame_extention_length &&
				strcmp(
					entry->d_name + length - savegame_extention_length,
					savegame_extention) == 0)
			{
				sprintf(fullname, "%s/%s", directoryname, entry->d_name);

				printf("Loading file %s\n", fullname);

				if (!load_savegame_from_filename(fullname, pperror))
				{
					free_savegame_information();
					return FALSE;
				}
				
				if (gameid == NULL)
				{
					gameid = strdup(game.id);
				}

				if (strcmp(game.id, gameid) == 0)
				{
					savegame_information_t* psavegame_information =
						(savegame_information_t*) malloc(sizeof(savegame_information_t));
	
					psavegame_information->filename = strdup(fullname);
					psavegame_information->year = game.year;
	
					savegame_information_list_insert_back(
						&savegame_information,
						psavegame_information);
				}
				else
				{
					printf(
						"Ignoring %s, it has ID %s, first game had ID %s\n",
						entry->d_name,
						game.id,
						gameid);
				}
			}
			else
			{
				printf("Ignoring %s\n", entry->d_name);
			}
		}
		else
		{
			closedir(dir);
			dir = NULL;
		}
	}

	free(gameid);

	if (savegame_information_list_size(&savegame_information) == 0)
	{
		g_set_error(
			pperror,
			0,
			0,
			"Could not find savegames in directory: %s",
			directoryname);
		return FALSE;
	}

	savegame_information_list_sort(
		&savegame_information,
		compare_savegame_information_t);

	return goto_turn(0, pperror);
}

bool load_savegame_from_filename(char* filename, GError** pperror)
{
	struct section_file savegame;

	if (!section_file_load_nodup(&savegame, filename))
	{
		g_set_error(
			pperror,
			0,
			0,
			"Couldn't load savefile: %s",
			filename);
		return FALSE;
	}

	bool result = load_savegame(&savegame, pperror);

	section_file_free(&savegame);

	return result;
}

void set_settings(engine_settings* psettings)
{
	g_settings = *psettings;
	
	draw_map();
}
