#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>

#include "engine.h"
#include "gui.h"

#include "config.h"

#include "map.h"

enum
{
	PIT_COL_ID = 0,
	PIT_COL_NAME,
	PIT_COL_POPULATION,
	PIT_COL_PRODUCTION,
	PIT_COL_ECONOMICS,
	PIT_NUM_COLS
};

GtkWidget*		load_savegame_button;
GtkWidget*		turn_label;

GtkWidget*		view_map_of_label;
GtkWidget*		view_map_of_combo_box;

GtkWidget*		main_window;
GtkWidget*		map_canvas;

GtkListStore*	player_information_liststore;
GtkTreeModel*	player_information_treemodel;
GtkWidget*		player_information_treeview;

GtkAdjustment*	phorizontal_adjustments;
GtkWidget*		phorizontal_scrollbar;
GtkAdjustment*	pvertical_adjustments;
GtkWidget*		pvertical_scrollbar;

GtkAdjustment*	pturn_adjustments;
GtkWidget*		pturn_scrollbar;

GdkPixmap* pwork_pixmap = NULL;

void center_view(int x, int y);
void center_view_on_tile(struct tile* ptile);

gboolean	event_delete(GtkWidget* widget, GdkEvent* event, gpointer data);
void		event_destroy(GtkWidget* widget, gpointer data);
void		event_load_savegame_button_clicked(GtkWidget *widget, gpointer data);
gboolean	event_map_button_release(GtkWidget* widget, GdkEventButton *event);
gboolean	event_map_canvas_configure(GtkWidget *widget, GdkEventConfigure *event);
gboolean	event_map_canvas_expose(GtkWidget *widget, GdkEventExpose *event);
void		event_player_information_selection_changed(GtkTreeSelection* pselection, gpointer userdata);
void		event_turn_scrollbar_changed(GtkWidget *widget, gpointer data);
void		event_view_changed(GtkWidget *widget, gpointer data);
void		event_view_world_of_changed(GtkWidget *widget, gpointer data);

void update_map_scrollbars(void);
void update_visible_map(void);

void center_view(int x, int y)
{
	if (x < 0)
	{
		if (topo_has_flag(TF_WRAPX))
		{
			// Make sure we can also draw rediculously small maps
			// in combination with _weird input
			while (x < 0) x += map.xsize;
		}
		else
		{
			x = 0;
		}
	}
	else if (x >= map.xsize)
	{
		if (topo_has_flag(TF_WRAPX))
		{
			while (x >= map.xsize) x -= map.xsize;
		}
		else
		{
			x = map.xsize - 1;
		}
	}

	if (y < 0)
	{
		if (topo_has_flag(TF_WRAPY))
		{
			// Make sure we can also draw rediculously small maps
			// in combination with _weird input
			while (y < 0) y += map.ysize;
		}
		else
		{
			y = 0;
		}
	}
	else if (y >= map.ysize)
	{
		if (topo_has_flag(TF_WRAPY))
		{
			while (y >= map.ysize) y -= map.ysize;
		}
		else
		{
			y = map.ysize - 1;
		}
	}
	
	phorizontal_adjustments->value = x;
	pvertical_adjustments->value = y;

	g_signal_emit_by_name(G_OBJECT(phorizontal_adjustments), "value_changed");
	g_signal_emit_by_name(G_OBJECT(pvertical_adjustments), "value_changed");
}

void center_view_on_tile(struct tile* ptile)
{
	center_view(ptile->x, ptile->y);
}

gboolean event_delete(GtkWidget* widget, GdkEvent* event, gpointer data)
{
    return FALSE;
}

void event_destroy(GtkWidget* widget, gpointer data)
{
    gtk_main_quit();
}

void event_load_savegame_button_clicked(GtkWidget *widget, gpointer data)
{
	int i;
	GError* perror = NULL;

	// 1) Get a directory from the user	
	GtkWidget *dialog;
	char *filename;
	
	dialog = gtk_file_chooser_dialog_new(
		"Choose a SaveGame Directory",
		GTK_WINDOW(main_window),
		GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER,
		GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
		GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
		NULL);
	
	gint result = gtk_dialog_run(GTK_DIALOG(dialog));
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
	
	gtk_widget_destroy(dialog);
	
	if (result != GTK_RESPONSE_ACCEPT)
	{
		g_free(filename);
		return;
	}

	// 2) Load the game
	if (!load_savegame_directory(filename, &perror))
	{
		dialog = gtk_message_dialog_new(
			GTK_WINDOW(main_window),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_CLOSE,
			"Error loading file '%s': %s",
			filename,
			perror->message);

		gtk_dialog_run(GTK_DIALOG(dialog));

		gtk_widget_destroy(dialog);

		return;
	}

	g_free(filename);

	// 3) Create new store and model for the player_information_treeview
	// One day, free the model when we enter the function ... g_object_unref(model);
	GtkTreeIter iter;

	player_information_liststore = gtk_list_store_new(
		PIT_NUM_COLS,
		G_TYPE_INT,
		G_TYPE_STRING,
		G_TYPE_STRING,
		G_TYPE_STRING,
		G_TYPE_STRING);

	for(i = 0; i < game.nplayers; i++)
	{
		gtk_list_store_append(player_information_liststore, &iter);
		gtk_list_store_set(
			player_information_liststore,
			&iter,
			PIT_COL_ID, i,
			PIT_COL_NAME, game.players[i].name,
			-1);
	}

	player_information_treemodel = GTK_TREE_MODEL(player_information_liststore);

	gtk_tree_view_set_model(
		GTK_TREE_VIEW(player_information_treeview),
		player_information_treemodel);

	// 4) Update "View map of" combobox names.
	gtk_list_store_clear(GTK_LIST_STORE(gtk_combo_box_get_model(GTK_COMBO_BOX(view_map_of_combo_box))));

	gtk_combo_box_append_text(GTK_COMBO_BOX(view_map_of_combo_box), "Entire map");

	for(i = 0; i < game.nplayers; i++)
	{
		gtk_combo_box_append_text(GTK_COMBO_BOX(view_map_of_combo_box), game.players[i].name);
	}

	gtk_combo_box_set_active(GTK_COMBO_BOX(view_map_of_combo_box), 0);

	// 5) Update adjustments for the turn scrollbar, this means the
	// function event_turn_scrollbar_changed will be called. This will update
	// some bits and pieces of the UI as well.
	pturn_adjustments->value = 0;
	pturn_adjustments->upper = get_turn_count();

	gtk_range_set_adjustment(GTK_RANGE(pturn_scrollbar), pturn_adjustments);
	g_signal_emit_by_name(G_OBJECT(pturn_adjustments), "changed");
	
	// 6) Update the map scrollbars. This will also redraw the map.
	update_map_scrollbars();
}

gboolean event_map_button_release(GtkWidget* widget, GdkEventButton *event)
{
	gint xsize, ysize;
	
	gdk_drawable_get_size(map_canvas->window, &xsize, &ysize);

	int current_x = (int) gtk_adjustment_get_value(phorizontal_adjustments);
	int current_y = (int) gtk_adjustment_get_value(pvertical_adjustments);

	current_x += (event->x - xsize / 2 - 15) / 30 + 0.5;
	current_y += (event->y - ysize / 2 - 15) / 30 + 0.5;

	center_view(current_x, current_y);
	
	return TRUE;
}

gboolean event_map_canvas_configure(GtkWidget *widget, GdkEventConfigure *event)
{
	if (pwork_pixmap != NULL)
	{
		g_object_unref(pwork_pixmap);
	}

	pwork_pixmap = gdk_pixmap_new(
		widget->window,
		widget->allocation.width,
		widget->allocation.height,
		-1);

	update_map_scrollbars();
	
	return TRUE;
}

gboolean event_map_canvas_expose(GtkWidget *widget, GdkEventExpose *event)
{
	update_visible_map();
	
	/* consider something like:
	gdk_draw_drawable(
		widget->window,
		widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		pixmap,
		event->area.x, event->area.y,
		event->area.x, event->area.y,
		event->area.width, event->area.height);
	*/
	
	return FALSE;
}

void event_player_information_selection_changed(GtkTreeSelection* pselection, gpointer userdata)
{
	GtkTreeModel*	pmodel;
	GtkTreeIter		iter;

	if (gtk_tree_selection_get_selected(pselection, &pmodel, &iter))
	{
		int player_id;

		gtk_tree_model_get(pmodel, &iter, PIT_COL_ID, &player_id, -1);

		int city_count = city_list_size(&game.players[player_id].cities);

		if (city_count == 0)
		{
			return;
		}

		struct city *pcity = find_palace(&game.players[player_id]);

		if (pcity == NULL)
		{
			pcity = city_list_get(&game.players[player_id].cities, 0);
		}

		center_view_on_tile(pcity->tile);
	}
}

void event_turn_scrollbar_changed(GtkWidget *widget, gpointer data)
{
	GError* perror = NULL;

	int turn = (int) gtk_adjustment_get_value(pturn_adjustments);

	if (!goto_turn(turn, &perror))
	{
		GtkWidget* dialog = gtk_message_dialog_new(
			GTK_WINDOW(main_window),
			GTK_DIALOG_DESTROY_WITH_PARENT,
			GTK_MESSAGE_ERROR,
			GTK_BUTTONS_CLOSE,
			"Error switching to turn %d: %s",
			turn,
			perror->message);

		gtk_dialog_run(GTK_DIALOG(dialog));

		gtk_widget_destroy(dialog);
	}

	// Update turn label
	char buffer[1024];

	sprintf(buffer, "Year: %d%s (Turn %d/%d)",
		abs(game.year),
		(game.year < 0) ? "BC" : "AD",
		turn + 1,
		get_turn_count());
	gtk_label_set_text(GTK_LABEL(turn_label), (const char*) buffer);

	// Update player treeview
	GtkTreeIter iter;
	int player_count = 0;
	char population_buffer[1024];
	char production_buffer[1024];
	char economics_buffer[1024];

	gboolean valid = gtk_tree_model_get_iter_first(
		player_information_treemodel,
		&iter);

	while (valid && player_count < game.nplayers)
	{
		sprintf(population_buffer, "%dK", game.players[player_count].score.population);
		sprintf(production_buffer, "%d", game.players[player_count].score.mfg);
		sprintf(economics_buffer, "%d", game.players[player_count].score.bnp);

		gtk_list_store_set(
			player_information_liststore,
			&iter,
			PIT_COL_POPULATION, population_buffer,
			PIT_COL_PRODUCTION, production_buffer,
			PIT_COL_ECONOMICS, economics_buffer,
			-1);

		valid = gtk_tree_model_iter_next(player_information_treemodel, &iter);
		player_count++;
    }

	// Update map
	update_visible_map();
}

void event_view_changed(GtkWidget *widget, gpointer data)
{
	update_visible_map();
}

void event_view_world_of_changed(GtkWidget *widget, gpointer data)
{
	int current_selection = gtk_combo_box_get_active(GTK_COMBO_BOX(view_map_of_combo_box));

	engine_settings settings;
	
	get_settings(&settings);

	if (current_selection == 0)
	{
		settings.current_player = SHOW_ALL_PLAYERS;
	}
	else
	{
		settings.current_player = current_selection - 1;
	}
	
	set_settings(&settings);
	
	update_visible_map();
}

GtkWidget* get_mapcanvas_widget(void)
{
	return map_canvas;
}

void gui_main()
{
	gtk_widget_show(main_window);

	gtk_main();
}

void initialize_gui()
{
	//////////////////////////////////////////////////////////////////////////
	// Main box, contains a panel on the left, and the map on the right

	GtkWidget* main_box = gtk_hbox_new(FALSE, 5);

	//////////////////////////////////////////////////////////////////////////
	// Left panel

	GtkWidget* left_box = gtk_vbox_new(FALSE, 5);

	// Load savegame button
	load_savegame_button = gtk_button_new_with_label("Load savegame");
	g_signal_connect(G_OBJECT(load_savegame_button), "clicked",
					 G_CALLBACK(event_load_savegame_button_clicked), NULL);
	gtk_box_pack_start(GTK_BOX(left_box), load_savegame_button, FALSE, FALSE, 0);
	gtk_widget_show(load_savegame_button);

	// Turn scrollbar
	pturn_adjustments = GTK_ADJUSTMENT(
		gtk_adjustment_new(0.0, 0.0, 1.0, 1.0, 1.0, 1.0));
	pturn_scrollbar = gtk_hscrollbar_new(pturn_adjustments);
	g_signal_connect(
		G_OBJECT(pturn_adjustments),
		"value_changed",
		G_CALLBACK(event_turn_scrollbar_changed),
		NULL);
	g_signal_connect(
		G_OBJECT(pturn_adjustments),
		"changed",
		G_CALLBACK(event_turn_scrollbar_changed),
		NULL);
	gtk_range_set_update_policy(
		GTK_RANGE(pturn_scrollbar),
		GTK_UPDATE_CONTINUOUS);
	gtk_box_pack_start(GTK_BOX(left_box), pturn_scrollbar, FALSE, FALSE, 0);
	gtk_widget_show(pturn_scrollbar);

	// Turn label
	turn_label = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(left_box), turn_label, FALSE, FALSE, 0);
	gtk_widget_show(turn_label);

	// "View map of" widgets
	view_map_of_label = gtk_label_new("View map of");
	gtk_box_pack_start(GTK_BOX(left_box), view_map_of_label, FALSE, FALSE, 0);
	gtk_widget_show(view_map_of_label);

	view_map_of_combo_box = gtk_combo_box_new_text();
	gtk_box_pack_start(GTK_BOX(left_box), view_map_of_combo_box, FALSE, FALSE, 0);
	gtk_widget_show(view_map_of_combo_box);

	g_signal_connect(G_OBJECT(view_map_of_combo_box), "changed",
					 G_CALLBACK(event_view_world_of_changed), NULL);

	// TreeView that displays the player information, the model is created
	// when the game is loaded.
	player_information_treeview = gtk_tree_view_new();

	GtkCellRenderer* renderer;

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(player_information_treeview),
		-1,
		"Name",
		renderer,
		"text",
		PIT_COL_NAME,
		NULL);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(player_information_treeview),
		-1,
		"Pop",
		renderer,
		"text",
		PIT_COL_POPULATION,
		NULL);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(player_information_treeview),
		-1,
		"Prod",
		renderer,
		"text",
		PIT_COL_PRODUCTION,
		NULL);

	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_insert_column_with_attributes(
		GTK_TREE_VIEW(player_information_treeview),
		-1,
		"Econ",
		renderer,
		"text",
		PIT_COL_ECONOMICS,
		NULL);

	GtkTreeSelection* pselection = gtk_tree_view_get_selection(
		GTK_TREE_VIEW(player_information_treeview));

	g_signal_connect(
		G_OBJECT(pselection),
		"changed",
		G_CALLBACK(event_player_information_selection_changed),
		NULL);

	gtk_box_pack_start(GTK_BOX(left_box), player_information_treeview, TRUE, TRUE, 0);
	gtk_widget_show(player_information_treeview);

	// Done, add left_box to main_box
	gtk_box_pack_start(GTK_BOX(main_box), left_box, FALSE, FALSE, 0);
	gtk_widget_show(left_box);

	//////////////////////////////////////////////////////////////////////////
	// Map canvas and scrollbars

	GtkWidget* map_table = gtk_table_new(2, 2, FALSE);

	// Map canvas
	map_canvas = gtk_drawing_area_new();

	gtk_widget_add_events(map_canvas,
		GDK_BUTTON_RELEASE_MASK |
		GDK_BUTTON_PRESS_MASK);

	g_signal_connect(G_OBJECT(map_canvas), "configure_event",
					 G_CALLBACK(event_map_canvas_configure), NULL);
	g_signal_connect(G_OBJECT(map_canvas), "expose_event",
					 G_CALLBACK(event_map_canvas_expose), NULL);
	g_signal_connect(G_OBJECT(map_canvas), "button_release_event",
					 G_CALLBACK(event_map_button_release), NULL);

   	gtk_table_attach(
		GTK_TABLE(map_table),
		map_canvas,
		0, 1, 0, 1,
		GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND,
		0, 0);
	gtk_widget_show(map_canvas);

	// Horizontal scrollbar
	phorizontal_adjustments = GTK_ADJUSTMENT(
		gtk_adjustment_new(0.0, 0.0, 1.0, 1.0, 1.0, 1.0));
	phorizontal_scrollbar = gtk_hscrollbar_new(phorizontal_adjustments);
	gtk_range_set_update_policy(
		GTK_RANGE(phorizontal_scrollbar),
		GTK_UPDATE_CONTINUOUS);
	g_signal_connect(
		G_OBJECT(phorizontal_adjustments),
		"value_changed",
		G_CALLBACK(event_view_changed),
		NULL);
	gtk_table_attach(
		GTK_TABLE(map_table),
		phorizontal_scrollbar,
		0, 1, 1, 2,
		GTK_FILL, GTK_FILL,
		0, 0);
	gtk_widget_show(phorizontal_scrollbar);

	// Vertical scrollbar
	pvertical_adjustments = GTK_ADJUSTMENT(
		gtk_adjustment_new(0.0, 0.0, 1.0, 1.0, 1.0, 1.0));
	pvertical_scrollbar = gtk_vscrollbar_new(pvertical_adjustments);
	gtk_range_set_update_policy(
		GTK_RANGE(pvertical_scrollbar),
		GTK_UPDATE_CONTINUOUS);
	g_signal_connect(
		G_OBJECT(pvertical_scrollbar),
		"value_changed",
		G_CALLBACK(event_view_changed),
		NULL);
	gtk_table_attach(
		GTK_TABLE(map_table),
		pvertical_scrollbar,
		1, 2, 0, 1,
		GTK_FILL, GTK_FILL,
		0, 0);
	gtk_widget_show(pvertical_scrollbar);

	// Done, add map_table to main_box
	gtk_box_pack_start(GTK_BOX(main_box), map_table, TRUE, TRUE, 0);
	gtk_widget_show(map_table);

	//////////////////////////////////////////////////////////////////////////
	// Actual window
	main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(main_window), "delete_event",
					 G_CALLBACK(event_delete), NULL);
	g_signal_connect(G_OBJECT(main_window), "destroy",
					 G_CALLBACK(event_destroy), NULL);
	gtk_window_set_default_size(GTK_WINDOW(main_window), 800, 600);
	gtk_container_set_border_width(GTK_CONTAINER(main_window), 10);
	gtk_container_add(GTK_CONTAINER(main_window), main_box);
	gtk_widget_show(main_box);
}

void update_map_scrollbars(void)
{
	phorizontal_adjustments->upper = map.xsize;
	pvertical_adjustments->upper = map.ysize;

	g_signal_emit_by_name(G_OBJECT(phorizontal_adjustments), "changed");
	g_signal_emit_by_name(G_OBJECT(pvertical_adjustments), "changed");
}

void update_visible_map(void)
{
	gint xsize, ysize;
	
	gdk_drawable_get_size(map_canvas->window, &xsize, &ysize);

	// Clear the old work pixmap
	gdk_draw_rectangle(
		pwork_pixmap,
		map_canvas->style->black_gc,
		TRUE,
		0,
		0,
		xsize,
		ysize);

	// Determine what to draw
	int x = (int) gtk_adjustment_get_value(phorizontal_adjustments);
	int y = (int) gtk_adjustment_get_value(pvertical_adjustments);

	// Draw the current map, if available
	if (is_game_loaded())
	{
		// HACK: ... bluntly copy the map a couple of times :p
		int xshift, yshift;
		int max_xshift = topo_has_flag(TF_WRAPX) ? 1 : 0;
		int max_yshift = topo_has_flag(TF_WRAPY) ? 1 : 0;

		for (xshift = -max_xshift; xshift <= max_xshift; xshift++)
		{
			for (yshift = -max_yshift; yshift <= max_yshift; yshift++)
			{
				gdk_draw_drawable(
					pwork_pixmap,
				    map_canvas->style->fg_gc[GTK_WIDGET_STATE(map_canvas)],
				    get_map(),
				    0,
				    0,
				    xsize / 2 - 15 - x * 30 - xshift * map.xsize * 30,
				    ysize / 2 - 15 - y * 30 - yshift * map.ysize * 30,
				    map.xsize * 30,
				    map.ysize * 30);
			}
		}
	}

	// Copy the work pixmap to the screen
	gdk_draw_drawable(
		map_canvas->window,
	    map_canvas->style->fg_gc[GTK_WIDGET_STATE(map_canvas)],
	    pwork_pixmap,
	    0, 0,
	    0, 0,
	    xsize, ysize);
}
